import re

import click


class RegexType(click.ParamType):
    name = "regex"

    def convert(self, value, param, ctx):
        try:
            return re.compile(value, re.IGNORECASE)
        except re.error as e:
            self.fail(
                "Regex error: " + str(e),
                param,
                ctx,
            )



import re
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional


@dataclass(order=True)
class Parameter:
    """Represent a MAVLink parameter."""

    name: str
    value: Any = None
    index: int = 0
    type: int = 10

    def __post_init__(self):
        # Convert the value according to the type:
        # https://mavlink.io/en/messages/common.html#MAV_PARAM_TYPE
        if self.value is not None:
            if self.is_float:
                self.value = float(self.value)
            else:
                self.value = int(self.value)
        self.name = self.name.upper()

    @classmethod
    def from_message(cls, message) -> "Parameter":
        return Parameter(
            name=message["param_id"],
            value=message["param_value"],
            index=message["param_index"],
            type=message["param_type"],
        )

    @property
    def is_float(self):
        """Return whether the value of this parameter is a float."""
        return self.type in (9, 10)

    def as_dict(self) -> Dict[str, Any]:
        return {
            "value": self.value,
            "index": self.index,
            "type": self.type,
        }


@dataclass()
class Backup:
    """Represent an entire backup file."""

    status: List[str] = field(default_factory=list)
    schema: int = 1
    timestamp: datetime = field(default_factory=datetime.now)
    parameters: Dict[str, Parameter] = field(default_factory=dict)

    @property
    def status_str(self) -> str:
        """Construct a status comment string from a list of status messages."""
        if not self.status:
            return ""
        return "// " + "\n// ".join(self.status) + "\n\n"

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Backup":
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["parameters"] = {
            name: Parameter(name=name, **data) for name, data in d["parameters"].items()
        }
        backup = cls(**d)
        return backup

    def as_dict(self) -> Dict[str, Any]:
        return {
            "schema": self.schema,
            "timestamp": self.timestamp.isoformat(),
            "status": self.status,
            "parameters": {
                name: param.as_dict() for name, param in sorted(self.parameters.items())
            },
        }

    def filter(self, regex: Optional[re.Pattern]) -> None:
        """
        Filter parameters based on a regex.

        Modifies the backup in-place.
        """
        if not regex:
            return

        self.parameters = {
            name: param for name, param in self.parameters.items() if regex.search(name)
        }



import sys
import time
from typing import Generator
from typing import List
from typing import Union

from pymavlink import mavutil

from .data_types import Parameter


FORBIDDEN_PARAMETERS = (
    "COMPASS_DEV_ID",
    "COMPASS_DEV_ID2",
    "COMPASS_DEV_ID3",
    "COMPASS_DEV_ID4",
    "COMPASS_DEV_ID5",
    "COMPASS_DEV_ID6",
    "COMPASS_DEV_ID7",
    "COMPASS_DEV_ID8",
    "COMPASS_MOTCT",
    "ESC_CALIBRATION",
    "FENCE_TOTAL",
    "H_SV_MAN",
    "H_SW2_H3_ENABLE",
    "H_SW_H3_ENABLE",
    "INS_GYROFFS_X",
    "INS_GYROFFS_Y",
    "INS_GYROFFS_Z",
    "MIS_TOTAL",
    "SYS_NUM_RESETS",
)


class CraftCommunication:
    def __init__(self, socket: str):
        self._conn = mavutil.mavlink_connection(socket)
        self._conn.wait_heartbeat()

    def reset_to_default(self):
        """Reset all parameters to their default values."""
        self._conn.mav.command_long_send(
            self._conn.target_system,
            self._conn.target_component,
            mavutil.mavlink.MAV_CMD_PREFLIGHT_STORAGE,
            0,
            2.0,
            0,
            0,
            0,
            0,
            0,
            0,
        )

    def reboot(self) -> None:
        """Reboot the craft."""
        self._conn.reboot_autopilot()

    def get_param(self, parameter: Parameter) -> Parameter:
        """Retrieve a parameter from a craft."""
        # Request the parameter.
        self._conn.mav.param_request_read_send(
            self._conn.target_system,
            self._conn.target_component,
            parameter.name.encode(),
            -1,
        )

        # Wait for the response.
        start = time.time()
        while True:
            if time.time() > start + 5:
                sys.exit(
                    f"Craft did not respond to a query about {parameter.name}, the parameter might not exist."
                )

            message = self._conn.recv_match(blocking=True).to_dict()
            if (
                message.get("mavpackettype", "") == "PARAM_VALUE"
                and message["param_id"] == parameter.name
            ):
                break
        return Parameter.from_message(message)

    def set_param(self, parameter: Parameter) -> None:
        """Set a parameter on a craft."""
        self._conn.mav.param_set_send(
            self._conn.target_system,
            self._conn.target_component,
            parameter.name.encode(),
            parameter.value,
            parameter.type,
        )

    def list_params(self) -> Generator[Union[List[str], int, Parameter], None, None]:
        """
        Read all the parameters from a craft.

        Yields the status as a list in first iteration, a count of all
        parameters in the second, and the parameters after that.
        """
        # From https://www.ardusub.com/developers/pymavlink.html.
        self._conn.mav.param_request_list_send(
            self._conn.target_system, self._conn.target_component, 1
        )
        status = []
        # This keeps track of the FSM state. If it's 0, we need to output the
        # status, 1 we need to output the count, and 2 is parameters until
        # we're done.
        fsm_state = 0
        while True:
            # Read all messages, as AP eventually freezes if we try to
            # ignore too many message types.
            message = self._conn.recv_match(blocking=True).to_dict()

            # The FC sends the status messages first and then starts sending
            # parameters, but it still might send unrelated status messages
            # after that. We're only interested in the first messages.
            if message.get("mavpackettype", "") == "STATUSTEXT" and fsm_state == 0:
                status.append(message["text"])
                continue
            elif message.get("mavpackettype", "") != "PARAM_VALUE":
                continue

            if fsm_state == 0:
                # We're done with status messages, yield the status.
                yield status
                fsm_state = 1
                continue
            elif fsm_state == 1:
                # This is the first parameter, yield the count and then the
                # parameter itself right away.
                yield message["param_count"]
                fsm_state = 2

            if (
                message["param_id"].startswith("STAT_")
                or message["param_id"] in FORBIDDEN_PARAMETERS
            ):
                # Ignore statistical parameters and parameters that should not
                # be changed manually.
                continue

            parameter = Parameter.from_message(message)
            yield parameter
            if parameter.index == message["param_count"] - 1:
                break



#!/usr/bin/env python3
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import TextIO

import click
import json5

from .craft_communication import CraftCommunication
from .data_types import Backup
from .data_types import Parameter
from .utils import RegexType


DEBUG = False


def debug(message: Any) -> None:
    """Print a message if DEBUG is True."""
    if DEBUG:
        click.echo(message)


def load_backup_from_file(backup_file: TextIO) -> Backup:
    """Load a list of parameters from a file."""
    data = json5.load(backup_file)
    return Backup.from_dict(data)


def save_backup_to_file(backup: Backup, outfile: TextIO) -> None:
    """Save a list of parameters to a file."""
    # Brace yourselves, we're going to perform some horrible hacks to
    # format the file for readability.

    # Write the status comment.
    outfile.write(backup.status_str)

    backup_dict = backup.as_dict()

    # Pop the parameters from the backup dict so we can write them later.
    parameters = backup_dict.pop("parameters")
    backup_dict["parameters"] = {}

    # Write the top-level backup.
    backup_str = json5.dumps(backup_dict, indent=2)

    param_str = "".join(
        (
            "    " + json5.dumps({name: data})[1:-1] + ",\n"
            for name, data in parameters.items()
        )
    )

    # Insert the param_str into the rest of the file and write it.
    outfile.write(backup_str[:-4] + "\n" + param_str + "  " + backup_str[-4:])


def read_backup_from_craft(socket: str) -> Backup:
    """Read all the parameters from a craft."""
    # From https://www.ardusub.com/developers/pymavlink.html.
    craft = CraftCommunication(socket)
    backup = Backup()
    param_generator = craft.list_params()
    backup.status = next(param_generator)  # type: ignore
    debug(backup.status)
    count = next(param_generator)
    for parameter in param_generator:
        debug(parameter)
        click.echo(f"- {str(parameter.index).rjust(4)} / {count}  \r", nl=False)  # type: ignore
        backup.parameters[parameter.name] = parameter  # type: ignore
    return backup


@click.group()
@click.option("--debug", is_flag=True, help="Print debug information.")
@click.version_option()
def cli(debug):
    global DEBUG
    DEBUG = debug


@cli.command(help="Compare a previous backup to the parameters on a craft.")
@click.argument("backup_file", type=click.File("r"))
@click.argument("backup_file2", type=click.File("r"), required=False)
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
@click.option(
    "-f",
    "--filter",
    type=RegexType(),
    help="Filter commands to process based on a regex.",
)
@click.option(
    "-b", "--only-backup", is_flag=True, help="Only show values in the backup file."
)
@click.option("-c", "--only-craft", is_flag=True, help="Only show values in the craft.")
def compare(
    backup_file, socket, filter, backup_file2=None, only_backup=False, only_craft=False
):
    if backup_file2:
        click.echo("Comparing parameters between two backups...")
        backup2 = load_backup_from_file(backup_file2)
        count = len(backup2.parameters)
        param_generator = backup2.parameters.values()

        click.echo(
            "Parameter name".ljust(24) + "Backup".rjust(24) + "Backup 2".rjust(24)
        )
    else:
        click.echo("Comparing parameters between a backup and a craft...")

        craft = CraftCommunication(socket)

        param_generator = craft.list_params()
        next(param_generator)
        count = next(param_generator)

        click.echo("Parameter name".ljust(24) + "Craft".rjust(24) + "Backup".rjust(24))

    click.echo("-" * (24 + 24 + 24))

    backup = load_backup_from_file(backup_file)
    backup.filter(filter)

    all_parameters = set(backup.parameters.keys())
    seen_parameters = set()

    # Print parameters on the craft.
    for parameter in param_generator:
        click.echo(f"- {str(parameter.index).rjust(4)} / {count}  \r", nl=False)
        # Ignore filtered parameters.
        if filter and not filter.search(parameter.name):
            continue

        if parameter.name in seen_parameters:
            # We've already seen this one.
            continue
        else:
            seen_parameters.add(parameter.name)

        if parameter.name in all_parameters:
            backup_value = backup.parameters[parameter.name].value
        else:
            if only_backup:
                # We only want to see parameters that exist in the file,
                # so skip this line.
                continue
            backup_value = "-"

        if parameter.value != backup_value:
            click.echo(
                parameter.name.ljust(24)
                + str(parameter.value).rjust(24)
                + str(backup_value).rjust(24)
            )

    if not only_craft:
        # Print parameters that were in the backup file but not the craft.
        for parameter_name in all_parameters - seen_parameters:
            click.echo(
                parameter_name.ljust(24)
                + "-".rjust(24)
                + str(backup.parameters[parameter_name].value).rjust(24)
            )

    click.echo("-" * (24 + 24 + 24))


@cli.command(help="Reset the craft parameters to their default values.")
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
def reset_to_defaults(socket):
    click.echo("Resetting all parameters to their default values...")
    craft = CraftCommunication(socket)
    craft.reset_to_default()
    click.echo("Rebooting...")
    craft.reboot()
    click.echo("Done.")


@cli.command(help="Reboot the controller.")
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
def reboot(socket):
    click.echo("Rebooting...")
    craft = CraftCommunication(socket)
    craft.reboot()
    click.echo("Done.")


@cli.command(help="Restore a previous backup to a craft.")
@click.argument("backup_file", type=click.File("r"))
@click.option(
    "-f",
    "--filter",
    type=RegexType(),
    help="Filter commands to process based on a regex.",
)
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
def restore(backup_file, socket, filter):
    click.echo("Restoring parameters...")
    craft = CraftCommunication(socket)
    backup = load_backup_from_file(backup_file)
    backup.filter(filter)
    for name, parameter in backup.parameters.items():
        click.echo(f"Setting {parameter.name} = {parameter.value}...")
        craft.set_param(parameter)
    click.echo("Done.")


@cli.command(help="Back up all the parameters from a craft to a file.")
@click.argument("craft_name")
@click.option(
    "-f",
    "--filter",
    type=RegexType(),
    help="Filter commands to process based on a regex.",
)
@click.option(
    "-o",
    "--outdir",
    default=".",
    show_default=True,
    type=click.Path(exists=True, file_okay=False, writable=True),
    help="Directory to write the backup file to.",
)
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
def backup(craft_name, socket, outdir, filter):
    click.echo(f"Reading parameters from {socket}...")
    backup = read_backup_from_craft(socket)
    click.echo("Writing to file...")
    filename = Path(outdir) / (
        f"{craft_name}_" f"{datetime.now().strftime('%Y-%m-%d_%H-%M')}" ".chute"
    )

    backup.filter(filter)

    # Write the actual backup.
    with filename.open("w") as outfile:
        save_backup_to_file(backup, outfile)
    click.echo("Done.")


@cli.command(help="Filter parameters by a regular expression.")
@click.argument("regex", type=RegexType())
@click.argument("backup_file", type=click.File("r"))
@click.argument("output_file", type=click.File("w"))
def filter(regex, backup_file, output_file):
    click.echo("Filtering based on regular expression...")
    backup = load_backup_from_file(backup_file)
    backup.filter(regex)
    save_backup_to_file(backup, output_file)
    click.echo("Done.")


@cli.command(help="Print parameters from either a backup file or a craft.")
@click.argument("backup_file", type=click.File("r"), required=False)
@click.option(
    "-f",
    "--filter",
    type=RegexType(),
    help="Filter commands to process based on a regex.",
)
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
def show(backup_file, filter, socket):
    click.echo("Retrieving parameters...")

    if backup_file:
        backup = load_backup_from_file(backup_file)
    else:
        backup = read_backup_from_craft(socket)

    backup.filter(filter)

    click.echo("Parameter name".ljust(24) + "Value".rjust(24))
    click.echo("-" * (24 + 24))
    for name, parameter in sorted(backup.parameters.items()):
        click.echo(name.ljust(24) + str(parameter.value).rjust(24))
    click.echo("-" * (24 + 24))


@cli.command(name="get", help="Get and print parameters.")
@click.argument("params", nargs=-1)
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
def get_params(params, socket):
    click.echo("Retrieving parameters...")
    craft = CraftCommunication(socket)

    click.echo("Parameter name".ljust(24) + "Value".rjust(24))
    click.echo("-" * (24 + 24))

    param_generator = craft.list_params()
    next(param_generator)

    for parameter in sorted(params):
        parameter = craft.get_param(Parameter(name=parameter))
        click.echo(parameter.name.ljust(24) + str(parameter.value).rjust(24))
    click.echo("-" * (24 + 24))


@cli.command(name="set", help="Set parameters.")
@click.argument("params", nargs=-1)
@click.option(
    "-s",
    "--socket",
    default="/dev/ttyACM0",
    show_default=True,
    help="Path to the USB socket.",
)
def set_params(params, socket):
    craft = CraftCommunication(socket)

    click.echo(
        "Parameter name".ljust(24) + "Old value".rjust(24) + "New value".rjust(24)
    )
    click.echo("-" * (24 * 3))

    for param in params:
        if "=" not in param:
            click.echo(f"Wrong parameter syntax: {param} (should be NAME=VALUE).")
            continue

        name, value = param.split("=")

        old_parameter = craft.get_param(Parameter(name=name))
        new_parameter = Parameter(name=name, value=value)
        click.echo(
            name.ljust(24)
            + str(old_parameter.value).rjust(24)
            + str(new_parameter.value).rjust(24)
        )

        craft.set_param(new_parameter)
    click.echo("-" * (24 * 3))
    click.echo("Done setting parameters.")


@cli.group(help="Convert a Parachute backup into another format.")
@click.option(
    "-f",
    "--filter",
    type=RegexType(),
    help="Filter commands to process based on a regex.",
)
@click.pass_context
def convert(ctx, filter):
    ctx.ensure_object(dict)

    ctx.obj["FILTER"] = filter


@convert.command(help='Convert into a QGroundControl ".params" file.')
@click.argument("backup_file", type=click.File("r"))
@click.argument("output_file", type=click.File("w"))
@click.pass_context
def qgc(ctx, backup_file, output_file):
    click.echo("Converting to a QGroundControl compatible file...")
    backup = load_backup_from_file(backup_file)
    backup.filter(ctx.obj["FILTER"])

    output_file.write("# Vehicle-Id Component-Id Name Value Type\n")
    output_file.writelines(
        f"1\t1\t{name}\t{parameter.value:.18g}\t{parameter.type}\n"
        for name, parameter in sorted(backup.parameters.items())
    )
    output_file.close()


@convert.command(help='Convert into a Mission Planner ".param" file.')
@click.argument("backup_file", type=click.File("r"))
@click.argument("output_file", type=click.File("w"))
@click.pass_context
def mp(ctx, backup_file, output_file):
    click.echo("Converting to a Mission Planner compatible file...")
    backup = load_backup_from_file(backup_file)
    backup.filter(ctx.obj["FILTER"])

    output_file.writelines(
        f"{name},{parameter.value:.9f}\n"
        if parameter.is_float
        else f"{name},{parameter.value}\n"
        for name, parameter in sorted(backup.parameters.items())
    )
    output_file.close()


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['parachute']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'json5>=0.9.5,<0.10.0',
 'pymavlink>=2.4.14,<3.0.0',
 'pyserial>=3.5,<4.0']

entry_points = \
{'console_scripts': ['parachute = parachute.cli:cli']}

setup_kwargs = {
    'name': 'parachute',
    'version': '0.3.5',
    'description': 'A lifeline for ArduPilot craft.',
    'long_description': 'Parachute\n=========\n\nParachute is a swiss army knife for ArduPilot settings. It helps you quickly and\neasily back up all your parameters to a file (and restore them). It also lets\nyou get/set them, filter them, diff them, restore them or convert them to\nparameter files compatible with Mission Planner/QGroundControl.\n\n\nInstallation\n------------\n\nInstalling Parachute is simple. You can use `pipx` (recommended):\n\n```\n$ pipx install parachute\n```\n\nOr `pip` (less recommended):\n\n```\n$ pip install parachute\n```\n\n\nUsage\n-----\n\nParachute is called like so:\n\n```\n$ parachute backup <craft name>\n```\n\nFor example:\n\n```\n$ parachute backup Mini-Drak\n```\n\nTo restore:\n\n```\n$ parachute restore backup.chute\n```\n\n\nConversion\n----------\n\nYou can also convert a Parachute file to a file compatible with Mission Planner or QGroundControl:\n\n```\n$ parachute convert qgc Mini-Drak_2021-03-02_02-29.chute Mini-Drak.params\n```\n\n\nFiltering\n---------\n\nYou can filter parameters based on a regular expression:\n\n```\n$ parachute filter "serial[123]_" Mini-Drak_2021-03-02_02-29.chute filtered.chute\n```\n\nSince all parameter names are uppercase, the regex is case-insensitive, for convenience.\n\nYou can also filter when converting:\n\n```\n$ parachute convert --filter=yaw mp Mini-Drak_2021-03-02_02-29.chute -\n```\n\n\nComparing\n---------\n\nYou can compare parameters in a backup with parameters on the craft:\n\n```\n$ parachute compare backup.chute\n```\n\n\nGetting/setting\n---------------\n\nYou can get and set parameters:\n\n```\n$ parachute get BATT_AMP_OFFSET BATT_AMP_PERVLT\n```\n\n```\n$ parachute set BATT_AMP_OFFSET=-0.0135 BATT_AMP_PERVLT=63.8826\n```\n',
    'author': 'Stavros Korokithakis',
    'author_email': 'hi@stavros.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/stavros/parachute',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
